#ifndef __RAYTRACER
#define __RAYTRACER

#include "RenderEngine.h"

#include "../common/scene.h"
#include "../common/Ray.h"
#include "../common/HitPoint.h"
#include "../Primitives/PrimitiveBase.h"
#include "../common/oldbuffer.h"
#include "../shader/shaders.h"
#include "../common/Vector3.h"
#include "../common/Camera.h"

#define STORE_HITDATA
//#define STORE_HITPOINT

class RayTracer : public RenderEngine
{
public:
	RayTracer(Canvas *c) : RenderEngine(c)
	{ }
	
	~RayTracer()
	{ }
	
	void render(SlimScene *scene);
	
	inline void aaclearCanvas()
	{
		flush_Color_buffer();
		clear_idata_buffer();
	}
	
	inline void calcRay(float x, float y, SlimScene *scene, Camera &camera)
	{
		intersect_data i_data;
		//intersect_data* i_data;
		
		//i_data = get_idata((int)(x+0.5), (int)(y+0.5));
		//i_data->start = vector3(CAMERA1->pos);
		//i_data->step=0;
		//i_data->pos.c[0] = x;
		//i_data->pos.c[1] = y;
		
		Ray r = camera.getEyeRay(x, y);
		HitPoint h;
		Color c;
		
		if( scene->hierarchy->rayIntersectAll(r, h) )
		{
			i_data.step = 0;
			i_data.obj = h.primitivePtr;
			i_data.proj = r.getDir();
			i_data.intersect = h.pos;
			h.primitivePtr->shader(&i_data, &c);
#ifdef STORE_HITPOINT
			scene->hitPoints->getIndex(x, y) = h.pos;
#endif
#ifdef STORE_HITDATA
			//scene->hitData->getIndex(x, y).distance = h.distance;
			scene->hitData->getIndex(x, y) = h;
#endif
		}
		else
		{
			i_data.proj = r.getDir();
			i_data.intersect = -main_scene->camera->look;
			scene->background->shader(&i_data, &c);
#ifdef STORE_HITDATA
			scene->hitData->getIndex(x, y).distance = 0.0f;
#endif
		}
		
		if ( h.skipShading )
		{
			c.r=h.r;
			c.g=h.g;
			c.b=h.b;
		}
		
		set_Color(x, y, &c);
	}
};


#endif
